local Utopia = Utopia
if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 34 $")

local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")
local gui = LibStub("AceGUI-3.0")
local engine
local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour
local rotate = Utopia.rotate
local band = bit.band

-- GetTreeDefinition
local GetTreeDefinition
do
	local treeDef
	function GetTreeDefinition()
		local tree = new()

		local function sortMe(a,b) return a.text < b.text end

		for category,info in pairs(Utopia.data) do
			tinsert(tree, {
				text = category,
				value = category,
				children = {}
			})
		end
		sort(tree, sortMe)

		for i,treeInfo in pairs(tree) do
			local category = treeInfo.text
			local info = Utopia.data[category]
			for spellName,data in pairs(info.spells) do
				tinsert(treeInfo.children, {
					text = data.name,
					value = data.name,
				})
			end
			sort(treeInfo.children, sortMe)
		end

		treeDef = tree
		function GetTreeDefinition()
			return treeDef
		end
		return treeDef
	end
end

-- OnTreeGroupSelected
local function OnTreeGroupSelected(widget, event, value)
	widget:ReleaseChildren()

	local category, spell = ("\001"):split(value)
	local data = Utopia.data[category]
	assert(data, format("Utopia.data['%s'] ~= nil", category or "nil"))
	local spellInfo = spell and data.spells[spell]

	local sf = gui:Create("ScrollFrame")
	sf.width = "fill"
	sf:SetLayout("flow")

	Utopia.detailsFrame:SetStatusText(L["DESC."..category])

	local desc = gui:Create("Label")
	desc:SetText(L["DESC."..category])
	desc.width = "fill"
	sf:AddChild(desc)

	if (not spellInfo) then
		return
	end

	local anyCanDo
	local players = new()
--Utopia:Print(format("Details for %s", spell or "nil"))
	for unit in Utopia:IterateRoster() do
		local _, class = UnitClass(unit)
		if (class == spellInfo.class) then
			if (spellInfo.requiredTalent) then
				if (Utopia:HasTalent(unit, spellInfo.requiredTalent)) then
--Utopia:Print(format("- %s has talent", UnitName(unit)))
					anyCanDo = true
				end
			else
--Utopia:Print("- Anyone can cast")
				anyCanDo = true
			end
			tinsert(players, unit)
		end
	end
	local function GetSortString(unit)
		local talent, improved = 0, 0
		local name = UnitName(unit)
		if (spellInfo.requiredTalent) then
			if (Utopia.talents[name]) then
				talent = (Utopia:HasTalent(name, spellInfo.requiredTalent) or 0) + 2
			else
				talent = 0
			end
		end
		if (spellInfo.improved) then
			if (Utopia.talents[name]) then
				improved = (Utopia:HasTalent(name, spellInfo.improved) or 0) + 2
			else
				improved = 0
			end
		end
		local ret = format("%d%d%s", 9 - talent, 9 - improved, name)
		return ret
	end

	sort(players, function(a, b)
		local sortA = GetSortString(a)
		local sortB = GetSortString(b)
		return sortA < sortB
	end)

	local amount = Utopia.getAmount(spellInfo)
	local maxAmount = Utopia.getAmount(spellInfo, nil, nil, nil, true)

	local amountNum = amount and tonumber(strmatch(amount, "(%d+)"))
	local maxAmountNum = maxAmount and tonumber(strmatch(maxAmount, "(%d+)"))
	local amountColour = "|cFF80FF80"
	if (amountNum and maxAmountNum) then
		if (amountNum < maxAmountNum) then
			amountColour = "|cFFE0E060"
		end
	end

--Utopia:Print(format("- amount = %q, anyCanDo = %s", amount or "nil", tostring(anyCanDo)))
	if (amount and amount ~= "") then
		local cat = Utopia.data[category]
		assert(cat)
		if (cat.attribute) then
			amount = amount .. " " .. cat.attribute
		end
		local label = gui:Create("Label")
		label.width = "fill"
		label:SetText(format(L["Maximum potential amount for group is %s%s"], anyCanDo and amountColour or "|cFFFF8080", anyCanDo and amount or L["Not available"]))
		sf:AddChild(label)
	end

	if (maxAmount and maxAmount ~= "") then
		local cat = Utopia.data[category]
		assert(cat)
		if (cat.attribute) then
			maxAmount = maxAmount .. " " .. cat.attribute
		end
		local label = gui:Create("Label")
		label.width = "fill"
		label:SetText(format(L["Maximum possible amount is %s%s"], "|cFF80FF80", maxAmount))
		sf:AddChild(label)
	end

	for i, unit in ipairs(players) do
		local canDo, improved
		if (spellInfo.requiredTalent) then
			canDo = Utopia:HasTalent(unit, spellInfo.requiredTalent)
		else
			canDo = true
		end
		if (spellInfo.improved) then
			improved = 9 - (Utopia:HasTalent(unit, spellInfo.improved) or 0)
		end

        local group = gui:Create("UtopiaCustomGroup")
        group:SetWidth(150)

		local name = gui:Create("UtopiaLabel")
		name:SetText(Utopia:ColourPlayer(unit))
		name:SetFont(GameFontNormal)
		name:SetWidth(100)
		group:AddChild(name)

		local label = gui:Create("Label")
		label.width = "fill"

		local amountLabel
		local gotTalents = Utopia.talents[UnitName(unit)]
		if (canDo) then
			if (spellInfo.improved) then
				if (not gotTalents) then
					group:SetBorderColor(0.4,0.4,0.4)
					label:SetText(format(L["Talents unscanned"]))
				else
					local imp = Utopia:HasTalent(unit, spellInfo.improved) or 0
					if (imp > 0) then
						if (imp < spellInfo.maxTalentPoints) then
							group:SetBorderColor(0.8,0.8,0.4)
							label:SetText(format(L["Can apply. Partially improved (|cFFFF8080%d|r of |cFF80FF80%d|r talent points)"], imp, spellInfo.maxTalentPoints))
						else
							group:SetBorderColor(0.4,0.8,0.4)
							label:SetText(format(L["Can apply. Improved"]))
						end
					else
						group:SetBorderColor(0.8,0.8,0.4)
						label:SetText(L["Can apply. Un-improved"])
					end
				end
			else
				--label:SetText("Can apply")
				group:SetBorderColor(0.4,0.8,0.4)

				if (spellInfo.maxTalentPoints and spellInfo.maxTalentPoints > 1) then
					if (not gotTalents) then
						group:SetBorderColor(0.4,0.4,0.4)
						label:SetText(format(L["Talents unscanned"]))
					else
						if (canDo < spellInfo.maxTalentPoints) then
							group:SetBorderColor(0.8,0.8,0.4)
							label:SetText(format(L["Can apply partially (|cFFFF8080%d|r of |cFF80FF80%d|r talent points)"], canDo, spellInfo.maxTalentPoints))
						end
					end
				end
			end

			amount = Utopia.getAmount(spellInfo, false, nil, engine.players[UnitGUID(unit)])
			if (amount and amount ~= "") then
				amountLabel = gui:Create("Label")
				amountLabel:SetText(amount)
			end
		else
			if (not gotTalents) then
				group:SetBorderColor(0.4,0.4,0.4)
				label:SetText(format(L["Talents unscanned"]))
			else
				group:SetBorderColor(0.8,0.4,0.4)
				label:SetText(L["Cannot apply this debuff"])
			end
		end

		group:AddChild(label)
		if (amountLabel) then
			group:AddChild(amountLabel)
		end

        sf:AddChild(group)
	end

	widget:AddChild(sf)
end

-- CreateDetailsFrame
function Utopia:CreateDetailsFrame()
	local f = gui:Create("Frame")
	self.detailsFrame = f
	f:SetTitle(self.title)
	f:SetStatusText("")
	f:SetWidth(600)
	f:SetHeight(400)
	f.width = "fill"
	f:SetLayout("flow")

	local tree = gui:Create("TreeGroup", 200)
	self.detailsTree = tree
	tree:EnableButtonTooltips(false)
	tree:SetLayout("fill")
	tree.width = "fill"
	tree.height = "fill"
	tree:SetTree(GetTreeDefinition())
	tree:SetCallback("OnGroupSelected", OnTreeGroupSelected, self)

	f:AddChild(tree)
	f:DoLayout()
end

-- firstActiveSpellForKey
local function firstActiveSpellForKey(key)
	info = Utopia.data[key]						-- DO NOT LOCALIZE, it's up there @ line -17

	for name,spellInfo in pairs(info.spells) do
		local canDo = Utopia.spellPotential and Utopia.spellPotential[name]
		if (not canDo) then
			if (spellInfo.class == "HUNTER") then
				canDo = Utopia.petPotential and Utopia.petPotential[name]
			end
		end

		if (canDo) then
			return name
		end
	end
end		

-- ToggleDetails
function Utopia:ToggleDetails(key)
	if (self.detailsFrame and self.detailsFrame:IsShown() and self.detailsTree.localstatus and subkey == self.detailsTree.localstatus.selected) then
		self.detailsFrame:Hide()
		self.detailsFrame:ReleaseChildren()
		self.detailsFrame = nil
		return
	end

	engine = UtopiaDebuffActivity

	if (not self.detailsFrame) then
		self:CreateDetailsFrame()
	end

	local subkey = firstActiveSpellForKey(key)
	if (not subkey) then
		local list = new()
		for name,spellInfo in pairs(info.spells) do
			tinsert(list, name)
		end
		sort(list)
		subkey = list[1]
		del(list)
	end

	self.detailsFrame:Show()
	self.detailsTree.localstatus.groups = {}
	self.detailsTree:SelectByValue(key)
	self.detailsTree:SelectByValue(subkey and key.."\001"..subkey or key)
end

do
	-- Copied from InlineGroup, but won't have title and allows for changing border colours
	local AceGUI = gui
	local Type = "UtopiaCustomGroup"
	local Version = 4
	
	local function OnAcquire(self)
		self:SetWidth(300)
		self:SetHeight(100)
		self.border:SetBackdropColor(0.1,0.1,0.1,0.5)
		self.border:SetBackdropBorderColor(0.4,0.4,0.4)
	end
	
	local function OnRelease(self)
		self.frame:ClearAllPoints()
		self.frame:Hide()
	end

	local PaneBackdrop  = {
		bgFile = "Interface\\ChatFrame\\ChatFrameBackground",
		edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 3, right = 3, top = 5, bottom = 3 }
	}
	
	local function SetTitle(self,title)
		self.titletext:SetText(title)
	end

	local function SetBorderColor(self, r, g, b)
		self.border:SetBackdropBorderColor(r, g, b)
	end

	local function SetColor(self, r, g, b)
		self.border:SetBackdropColor(r, g, b)
	end

	local function LayoutFinished(self, width, height)
		self:SetHeight((height or 0) + 25)
	end
	
	local function OnWidthSet(self, width)
		local content = self.content
		local contentwidth = width - 20
		if contentwidth < 0 then
			contentwidth = 0
		end
		content:SetWidth(contentwidth)
		content.width = contentwidth
	end
	
	
	local function OnHeightSet(self, height)
		local content = self.content
		local contentheight = height - 20
		if contentheight < 0 then
			contentheight = 0
		end
		content:SetHeight(contentheight)
		content.height = contentheight
	end
	
	local function Constructor()
		local frame = CreateFrame("Frame",nil,UIParent)
		local self = {}
		self.type = Type

		self.OnRelease = OnRelease
		self.OnAcquire = OnAcquire
		self.frame = frame
		self.LayoutFinished = LayoutFinished
		self.OnWidthSet = OnWidthSet
		self.OnHeightSet = OnHeightSet
		self.SetBorderColor = SetBorderColor
		self.SetColor = SetColor

		frame.obj = self

		frame:SetHeight(100)
		frame:SetWidth(100)
		frame:SetFrameStrata("FULLSCREEN_DIALOG")

		local border = CreateFrame("Frame",nil,frame)
		self.border = border
		border:SetPoint("TOPLEFT",frame,"TOPLEFT",3,-3)
		border:SetPoint("BOTTOMRIGHT",frame,"BOTTOMRIGHT",-3,3)

		border:SetBackdrop(PaneBackdrop)

		--Container Support
		local content = CreateFrame("Frame",nil,border)
		self.content = content
		content.obj = self
		content:SetPoint("TOPLEFT",border,"TOPLEFT",10,-10)
		content:SetPoint("BOTTOMRIGHT",border,"BOTTOMRIGHT",-10,10)

		AceGUI:RegisterAsContainer(self)
		return self
	end

	AceGUI:RegisterWidgetType(Type,Constructor,Version)
end

do
	local AceGUI = gui
	local Type = "UtopiaLabel"
	local Version = 8

	local function OnAcquire(self)
		self:SetText("")
		self:SetColor()
	end

	local function OnRelease(self)
		self.frame:ClearAllPoints()
		self.frame:Hide()
	end

	local function SetText(self, text)
		self.label:SetText(text or "")
		self.frame:SetHeight(self.label:GetStringHeight())
	end

	local function SetFont(self, fontObject)
		self.label:SetFontObject(fontObject)
		self.frame:SetHeight(self.label:GetStringHeight())
	end

	local function SetColor(self, r, g, b)
		if not (r and g and b) then
			r, g, b = 1, 1, 1
		end
		self.label:SetVertexColor(r, g, b)
	end

	local function Constructor()
		local frame = CreateFrame("Frame",nil,UIParent)
		local self = {}
		self.type = Type

		self.OnRelease = OnRelease
		self.OnAcquire = OnAcquire
		self.SetText = SetText
		self.SetFont = SetFont
		self.SetColor = SetColor
		self.frame = frame
		frame.obj = self

		frame:SetHeight(18)
		frame:SetWidth(200)
		local label = frame:CreateFontString(nil,"BACKGROUND","GameFontHighlightSmall")
		label:SetPoint("TOPLEFT",frame,"TOPLEFT",0,0)
		label:SetWidth(200)
		label:SetJustifyH("LEFT")
		label:SetJustifyV("TOP")
		self.label = label

		AceGUI:RegisterAsWidget(self)
		return self
	end

	AceGUI:RegisterWidgetType(Type,Constructor,Version)
end
