if (not Utopia) then
	return
end
Utopia:UpdateVersion("$Revision: 34 $")

local new, del, copy, deepDel
do
	--local z = ZOMGBuffs or Absorbaloth or PlayerRatings
	--if (z and z.new and z.del and z.copy and z.deepDel) then
	--	new, del, copy, deepDel = z.new, z.del, z.copy, z.deepDel
	--else
		local next, select, pairs, type = next, select, pairs, type
		local list = setmetatable({},{__mode='k'})
		function new(...)
			local t = next(list)
			if t then
--@debug@
				assert(not next(t))
--@no-debug@
				list[t] = nil
				for i = 1, select('#', ...) do
					t[i] = select(i, ...)
				end
				return t
			else
				return {...}
			end
		end
		function del(t)
			if (t) then
				setmetatable(t, nil)
				wipe(t)
				t[''] = true
				t[''] = nil
				list[t] = true
--@debug@
assert(not next(t))
--@no-debug@
			end
		end
		function deepDel(t)
			if (t) then
				setmetatable(t, nil)
				for k,v in pairs(t) do
					if type(v) == "table" then
						deepDel(v)
					end
					t[k] = nil
				end
				t[''] = true
				t[''] = nil
				list[t] = true
--@debug@
assert(not next(t))
--@no-debug@
			end
		end
		function copy(old)
			if (not old) then
				return
			end
			local n = new()
			for k,v in pairs(old) do
				if (type(v) == "table") then
					n[k] = copy(v)
				else
					n[k] = v
				end
			end
			setmetatable(n, getmetatable(old))
			return n
		end
	--end
end

Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel = new, del, copy, deepDel

do
	local format, min, max = string.format, math.min, math.max
	local r, g, b, sr, sg, sb
	local function push(c)
		local ret = format("|cFF%02X%02X%02X%s", r * 255, g * 255, b * 255, c)
		r = max(0, min(1, r + sr))
		g = max(0, min(1, g + sg))
		b = max(0, min(1, b + sb))
		return ret
	end

	function Utopia.Gradient(str, r1, g1, b1, r2, g2, b2)
		r, g, b = r1, g1, b1
		sr = (r2 - r1) / (#str - 1)
		sg = (g2 - g1) / (#str - 1)
		sb = (b2 - b1) / (#str - 1)
		return str:gsub(".", push)
	end
end

do
	local UnitName, format = UnitName, format
	function Utopia.UnitFullName(unit)
		local name, server = UnitName(unit)
		if (server and server ~= "") then
			return format("%s-%s", name, server)
		end
		return name
	end
end

do
	local classColours = {}
	for class, c in pairs(RAID_CLASS_COLORS) do
		classColours[class] = format("|cFF%02X%02X%02X", c.r * 255, c.g * 255, c.b * 255)
	end

	function Utopia.ClassColour(class)
		return classColours[class] or "|cFF808080"
	end

	function Utopia:ColourPlayer(unit)
		local _, class = UnitClass(unit)
		return format("%s%s|r", classColours[class], UnitName(unit))
	end
end

function Utopia.SmoothColour(percentage)
	local r, g
	if (percentage < 0.5) then
		g = min(1, max(0, 2*percentage))
		r = 1
	else
		g = 1
		r = min(1, max(0, 2*(1 - percentage)))
	end
	return r, g, 0
end

-- rotate
-- Positive angle is clockwise rotation
function Utopia.rotate(angle)
	local hcos, hsin = 0.5 * cos(angle), 0.5 * sin(angle)
	local mhcos, mhsin = -hcos, -hsin
	local ULx, ULy, LLx, LLy = hcos - mhsin, mhsin + hcos, hcos - hsin, mhsin + mhcos
	local URx, URy, LRx, LRy = mhcos - mhsin, hsin + hcos, mhcos - hsin, hsin + mhcos
	return ULx+0.5, ULy+0.5, LLx+0.5, LLy+0.5, URx+0.5, URy+0.5, LRx+0.5, LRy+0.5
end
